import base64
import asyncio
import traceback
import urllib.parse
import typing
import json
import logging
from dataclasses import dataclass

from mangum.lifespan import Lifespan
from mangum.utils import get_logger, make_response
from mangum.types import ASGIApp
from mangum.asgi import ASGIHTTPCycle, ASGIWebSocketCycle
from mangum.exceptions import ASGIWebSocketCycleException
from mangum.connections import ConnectionTable


@dataclass
class Mangum:

    app: ASGIApp
    debug: bool = False
    enable_lifespan: bool = True
    logger: logging.Logger = get_logger()

    def __post_init__(self) -> None:
        if self.enable_lifespan:
            loop = asyncio.get_event_loop()
            self.lifespan = Lifespan(self.app, logger=self.logger)
            loop.create_task(self.lifespan.run())
            loop.run_until_complete(self.lifespan.wait_startup())

    def __call__(
        self, event: typing.Dict[str, str], context: typing.Dict[str, str]
    ) -> typing.Dict[str, str]:
        try:
            response = self.handler(event, context)
        except Exception as exc:
            if self.debug:
                content = traceback.format_exc()
                return make_response(content, status_code=500)
            raise exc
        else:
            return response

    def get_server_and_client(
        self, event: typing.Dict[str, str]
    ) -> typing.Tuple:  # pragma: no cover
        """
        Parse the server and client keys for the scope definition, if possible.
        """
        client_addr = event["requestContext"].get("identity", {}).get("sourceIp", None)
        client = (client_addr, 0)
        server_addr = event["headers"].get("Host", None)
        if server_addr is not None:
            if ":" not in server_addr:
                server_port = 80
            else:
                server_port = int(server_addr.split(":")[1])

            server = (server_addr, server_port)
        else:
            server = None
        return server, client

    def handle_http(
        self, event: typing.Dict[str, str], context: typing.Dict[str, str]
    ) -> typing.Dict[str, str]:
        server, client = self.get_server_and_client(event)
        headers = [
            [k.lower().encode(), v.encode()] for k, v in event["headers"].items()
        ]
        query_string_params = event["queryStringParameters"]
        query_string = (
            urllib.parse.urlencode(query_string_params).encode()
            if query_string_params
            else b""
        )
        scope = {
            "type": "http",
            "http_version": "1.1",
            "method": event["httpMethod"],
            "headers": headers,
            "path": urllib.parse.unquote(event["path"]),
            "raw_path": None,
            "root_path": event["requestContext"]["stage"],
            "scheme": event["headers"].get("X-Forwarded-Proto", "https"),
            "query_string": query_string,
            "server": server,
            "client": client,
        }
        binary = event.get("isBase64Encoded", False)
        body = event["body"] or b""
        if binary:
            body = base64.b64decode(body)
        elif not isinstance(body, bytes):
            body = body.encode()
        asgi_cycle = ASGIHTTPCycle(scope, binary=binary)
        asgi_cycle.put_message(
            {"type": "http.request", "body": body, "more_body": False}
        )
        response = asgi_cycle(self.app)
        return response

    def handle_ws(
        self, event: typing.Dict[str, str], context: typing.Dict[str, str]
    ) -> typing.Dict[str, str]:
        request_context = event["requestContext"]
        connection_id = request_context.get("connectionId")
        domain_name = request_context.get("domainName")
        stage = request_context.get("stage")
        event_type = request_context["eventType"]
        endpoint_url = f"https://{domain_name}/{stage}"

        if event_type == "CONNECT":
            server, client = self.get_server_and_client(event)
            headers = event["headers"]
            root_path = event["requestContext"]["stage"]
            scope = {
                "type": "websocket",
                "path": "/ws",
                "headers": headers,
                "raw_path": None,
                "root_path": root_path,
                "scheme": event["headers"].get("X-Forwarded-Proto", "wss"),
                "query_string": "",
                "server": server,
                "client": client,
            }
            connection_table = ConnectionTable()
            status_code = connection_table.update_item(
                connection_id, scope=json.dumps(scope)
            )
            if status_code != 200:  # pragma: no cover
                # TODO: Improve error handling
                return make_response("Error", status_code=500)
            return make_response("OK", status_code=200)

        elif event_type == "MESSAGE":
            event_body = json.loads(event["body"])
            event_data = event_body["data"] or ""

            connection_table = ConnectionTable()
            item = connection_table.get_item(connection_id)
            if not item:  # pragma: no cover
                # TODO: Improve error handling
                return make_response("Error", status_code=500)

            # Retrieve and deserialize the scope entry created in the connect event for
            # the current connection.
            scope = json.loads(item["scope"])

            # Ensure the scope definitions comply with the ASGI spec.
            query_string = scope["query_string"]
            headers = scope["headers"]
            headers = [[k.encode(), v.encode()] for k, v in headers.items()]
            query_string = query_string.encode()
            scope.update({"headers": headers, "query_string": query_string})

            asgi_cycle = ASGIWebSocketCycle(
                scope,
                endpoint_url=endpoint_url,
                connection_id=connection_id,
                connection_table=connection_table,
            )
            asgi_cycle.put_message({"type": "websocket.connect"})
            asgi_cycle.put_message(
                {
                    "type": "websocket.receive",
                    "path": "/ws",
                    "bytes": None,
                    "text": event_data,
                }
            )
            try:
                asgi_cycle(self.app)
            except ASGIWebSocketCycleException:  # pragma: no cover
                # TODO: Improve error handling
                return make_response("Error", status_code=500)
            return make_response("OK", status_code=200)

        elif event_type == "DISCONNECT":
            connection_table = ConnectionTable()
            status_code = connection_table.delete_item(connection_id)
            if status_code != 200:  # pragma: no cover
                # TODO: Improve error handling
                return make_response("WebSocket disconnect error.", status_code=500)
            return make_response("OK", status_code=200)

    def handler(
        self, event: typing.Dict[str, str], context: typing.Dict[str, str]
    ) -> typing.Dict[str, str]:
        if "httpMethod" in event:
            response = self.handle_http(event, context)
        else:
            response = self.handle_ws(event, context)
        if self.enable_lifespan:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self.lifespan.wait_shutdown())
        return response



class ASGIWebSocketCycleException(Exception):
    """Raise when an exception occurs within an ASGI websocket cycle"""


class ConnectionTableException(Exception):
    """Raise when an exception occurs accessing data in the connection table"""



import logging


def make_response(content: str, status_code: int = 500) -> dict:
    return {
        "statusCode": status_code,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "body": content,
    }


def get_logger() -> logging.Logger:
    logging.basicConfig(
        format="[%(asctime)s] %(message)s",
        level=logging.INFO,
        datefmt="%d-%b-%y %H:%M:%S",
    )
    logger = logging.getLogger("mangum")
    logger.setLevel(logging.INFO)
    return logger



import typing
import os

import boto3
import botocore
from boto3.dynamodb.conditions import Attr

from mangum.exceptions import ConnectionTableException


class ConnectionTable:
    """
    Represents a DynamoDB resource that contains the WebSocket connection table.
    """

    def __init__(self) -> None:
        dynamodb = boto3.resource("dynamodb", region_name=os.environ["REGION_NAME"])
        self.table = dynamodb.Table(os.environ["TABLE_NAME"])

    def get_item(self, connection_id: str) -> typing.Union[typing.Dict, None]:
        """
        Retrieve an item in the connection table by id.
        """
        item = self.table.get_item(Key={"connectionId": connection_id}).get(
            "Item", None
        )
        return item

    def update_item(self, connection_id: str, **kwargs) -> int:
        """
        Update an item in the connection table by id.
        """
        result = self.table.put_item(Item={**{"connectionId": connection_id}, **kwargs})
        return result.get("ResponseMetadata", {}).get("HTTPStatusCode")

    def delete_item(self, connection_id: str) -> int:
        """
        Delete an item in the connection table by id.
        """
        result = self.table.delete_item(Key={"connectionId": connection_id})
        return result.get("ResponseMetadata", {}).get("HTTPStatusCode")

    def get_group_items(
        self, group: str
    ) -> typing.Union[typing.List[typing.Dict], None]:
        """
        Retrieve a list of items in the connection table by group.
        """
        items = self.table.scan(
            ProjectionExpression="connectionId",
            FilterExpression=Attr("groups").contains(group),
        ).get("Items", None)
        return items

    def send_data(
        self, items: typing.List[typing.Dict], data: str
    ) -> None:  # pragma: no cover
        """
        Send data to one or more items in the connection table.
        """
        apigw_management = boto3.client(
            "apigatewaymanagementapi", endpoint_url=self.endpoint_url
        )
        for item in items:
            try:
                apigw_management.post_to_connection(
                    ConnectionId=item["connectionId"], Data=data
                )
            except botocore.exceptions.ClientError as exc:
                status_code = exc.response.get("ResponseMetadata", {}).get(
                    "HTTPStatusCode"
                )
                if status_code == 410:
                    # Delete stale connection
                    self.connection_table.delete_item(
                        Key={"connectionId": item["connectionId"]}
                    )
                else:
                    raise ConnectionTableException("Connection does not exist")



import typing

ASGIScope = typing.Dict[str, typing.Any]
ASGIMessage = typing.Dict[str, typing.Any]
ASGIReceive = typing.Callable[[], typing.Awaitable[ASGIMessage]]
ASGISend = typing.Callable[[ASGIMessage], typing.Awaitable[None]]
ASGIApp = typing.Callable[[ASGIScope, ASGIReceive, ASGISend], typing.Awaitable[None]]



import enum
import asyncio
import base64
from dataclasses import dataclass, field

from mangum.connections import ConnectionTable
from mangum.types import ASGIScope, ASGIMessage, ASGIApp
from mangum.exceptions import ASGIWebSocketCycleException


class ASGICycleState(enum.Enum):
    REQUEST = enum.auto()
    RESPONSE = enum.auto()


@dataclass
class ASGICycle:

    scope: ASGIScope
    state: ASGICycleState = ASGICycleState.REQUEST
    binary: bool = False
    response: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.loop = asyncio.get_event_loop()
        self.app_queue = asyncio.Queue(loop=self.loop)

    def __call__(self, app: ASGIApp):
        asgi_instance = app(self.scope, self.asgi_receive, self.asgi_send)
        asgi_task = self.loop.create_task(asgi_instance)
        self.loop.run_until_complete(asgi_task)
        return self.response

    async def asgi_receive(self) -> dict:
        message = await self.app_queue.get()
        return message

    def put_message(self, message: ASGIMessage) -> None:
        self.app_queue.put_nowait(message)


@dataclass
class ASGIHTTPCycle(ASGICycle):

    body: bytes = b""

    async def asgi_send(self, message: ASGIMessage) -> None:
        if self.state is ASGICycleState.REQUEST:
            if message["type"] != "http.response.start":
                raise RuntimeError(
                    f"Expected 'http.response.start', received: {message['type']}"
                )

            status_code = message["status"]
            headers = {k: v for k, v in message.get("headers", [])}
            self.response["statusCode"] = status_code
            self.response["isBase64Encoded"] = self.binary
            self.response["headers"] = {
                k.decode(): v.decode() for k, v in headers.items()
            }
            self.state = ASGICycleState.RESPONSE

        elif self.state is ASGICycleState.RESPONSE:
            if message["type"] != "http.response.body":
                raise RuntimeError(
                    f"Expected 'http.response.body', received: {message['type']}"
                )

            body = message.get("body", b"")
            more_body = message.get("more_body", False)

            # The body must be completely read before returning the response.
            self.body += body

            if not more_body:
                body = self.body
                if self.binary:
                    body = base64.b64encode(body)
                self.response["body"] = body.decode()
                self.put_message({"type": "http.disconnect"})


@dataclass
class ASGIWebSocketCycle(ASGICycle):

    endpoint_url: str = None
    connection_id: str = None
    connection_table: ConnectionTable = None

    async def asgi_send(self, message: ASGIMessage) -> None:
        if self.state is ASGICycleState.REQUEST:
            if message["type"] in ("websocket.accept", "websocket.close"):
                self.state = ASGICycleState.RESPONSE
            else:
                raise RuntimeError(
                    f"Expected 'websocket.accept' or 'websocket.close', received: {message['type']}"
                )
        else:
            data = message.get("text")
            if message["type"] == "websocket.send":
                group = message.get("group", None)
                self.send_data(data=data, group=group)

    def send_data(self, *, data: str, group: str = None) -> None:  # pragma: no cover
        """
        Send a data message to a client or group of clients using the connection table.
        """
        item = self.connection_table.get_item(self.connection_id)
        if group:
            # Retrieve the existing groups for the current connection, or create a new
            # groups entry if one does not exist.
            groups = item.get("groups", [])
            if group not in groups:
                # Ensure the group specified in the message is included.
                groups.append(group)
                result = self.connection_table.update_item(
                    self.connection_id, groups=groups
                )
                status_code = result.get("ResponseMetadata", {}).get("HTTPStatusCode")
                if status_code != 200:
                    raise ASGIWebSocketCycleException("Error updating groups")

            # Retrieve all items associated with the current group.
            items = self.connection_table.get_group_items(group)
            if items is None:
                raise ASGIWebSocketCycleException("No connections found")
        else:
            # Single send, add the current item to a list to be iterated by the
            # connection table.
            items = [item]
        self.connection_table.send_data(items, data=data)



import typing
import logging
import asyncio
from dataclasses import dataclass


@dataclass
class Lifespan:

    app: typing.Any
    logger: logging.Logger
    startup_event: asyncio.Event = asyncio.Event()
    shutdown_event: asyncio.Event = asyncio.Event()
    app_queue: asyncio.Queue = asyncio.Queue()

    async def run(self):
        try:
            await self.app({"type": "lifespan"}, self.receive, self.send)
        except BaseException as exc:  # pragma: no cover
            self.logger.error(f"Exception in 'lifespan' protocol: {exc}")
        finally:
            self.startup_event.set()
            self.shutdown_event.set()

    async def send(self, message: dict) -> None:
        if message["type"] == "lifespan.startup.complete":
            self.startup_event.set()
        elif message["type"] == "lifespan.shutdown.complete":
            self.shutdown_event.set()
        else:  # pragma: no cover
            raise RuntimeError(
                f"Expected lifespan message type, received: {message['type']}"
            )

    async def receive(self) -> dict:
        message = await self.app_queue.get()
        return message

    async def wait_startup(self):
        self.logger.info("Waiting for application startup.")
        await self.app_queue.put({"type": "lifespan.startup"})
        await self.startup_event.wait()

    async def wait_shutdown(self):
        self.logger.info("Waiting for application shutdown.")
        await self.app_queue.put({"type": "lifespan.shutdown"})
        await self.shutdown_event.wait()
